import torch
import torch.nn as nn
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
import numpy as np

import matgl
from matgl.ext.ase import PESCalculator


class MatGLLoss(nn.Module):
    """
    Uses MatGL’s pre‑trained PES model to compute forces & energy for a displaced cell.
    """

    def __init__(self, model_name: str = "M3GNet-MP-2021.2.8-PES"):
        super().__init__()
        # load the MatGL PES model
        self.pot = matgl.load_model(model_name)
        # adaptor to convert pymatgen to ASE
        self.adaptor = AseAtomsAdaptor()

    def predict_force_energy(self, pos_flat: torch.Tensor, structure: Structure):
        """
        Given a flat tensor of displacements (shape 3N) and a pymatgen Structure,
        build an ASE Atoms at coords = structure.cart_coords + displacement,
        attach the PESCalculator, and return forces (flat 3N) and energy.
        """
        # 1) reshape pos_flat → (N, 3) numpy array
        disp = pos_flat.detach().cpu().numpy().reshape(-1, 3)

        # 2) get the base (perturbed) coords from the Structure
        base_coords = np.array(structure.cart_coords)  # shape (N,3)

        # 3) compute the new absolute coords
        coords = base_coords + disp  # shape (N,3)

        # 4) build ASE Atoms from the pymatgen Structure
        atoms = AseAtomsAdaptor.get_atoms(structure)

        # 5) update the ASE Atoms’ positions to your new coords
        atoms.positions = coords

        # 6) attach the calculator correctly
        calc = PESCalculator(potential=self.pot)
        atoms.calc = calc

        # 7) evaluate forces & energy
        forces = atoms.get_forces().astype(np.float32).reshape(-1)
        energy = float(atoms.get_potential_energy())

        # 8) convert back to torch
        f_tensor = torch.as_tensor(forces, dtype=torch.float32, device=pos_flat.device)
        e_tensor = torch.as_tensor(energy, dtype=torch.float32, device=pos_flat.device)
        return f_tensor, e_tensor

    def compute_fd_hessian(self, force_fn, disp0, eps=1e-3):
        """
        Placeholder while we figure out how to use torch's native hessian function
        """
        n = disp0.numel()
        H = torch.zeros(n, n, device=disp0.device)
        for j in range(n):
            e = torch.zeros_like(disp0)
            e[j] = eps
            f_plus, _ = force_fn(disp0 + e)
            f_minus, _ = force_fn(disp0 - e)
            H[:, j] = (f_plus - f_minus) / (2 * eps)
        return H

    def forward(
        self, input: torch.Tensor, structure: Structure, classifier
    ) -> torch.Tensor:
        """
        Compute loss between predicted displacements and the “true” displacement

        Parameters
        ----------
        input: torch.Tensor, shape (3*N,)
          The predicted displacement vector
        structure: Structure
          The perturbed pymatgen structure used to compute forces and Hessian
        classifier: Callable
          A loss function, e.g. nn.MSELoss()

        Returns
        -------
        loss: torch.Tensor
        """
        # build zero‐displacement vector with grad
        disp0 = torch.zeros_like(
            input, device=input.device, requires_grad=True
        )  # (3*N,)

        # get forces & energy at the perturbed geometry (disp0 = 0 → new_coords = structure.cart_coords)
        f0, e0 = self.predict_force_energy(disp0, structure)

        # Hessian in displacement space
        H = self.compute_fd_hessian(
            lambda d: self.predict_force_energy(d, structure), disp0
        )

        # Ideally we are able to run this
        # H  = torch.autograd.functional.hessian(lambda d: self.predict_force_energy(d, structure), disp0)

        # regularize & solve for δx: (H + λI) δx = –F₀
        diag_mean = H.diagonal().abs().mean()
        ridge_val = 1e-1 * diag_mean
        ridge = ridge_val * torch.eye(H.size(0), device=H.device)
        H_reg = H + ridge

        try:
            delta_x = torch.linalg.solve(H_reg, -f0)
        except RuntimeError:
            # fallback to least‑squares if still singular
            sol = torch.linalg.lstsq(H_reg, (-f0).unsqueeze(-1))
            delta_x = sol.solution.squeeze(-1)

        # reshape to match network’s output
        actual_disp = delta_x.view_as(input)

        print("Step norm:", actual_disp.norm().item())
        return classifier(input, actual_disp)



import torch
import torch.nn as nn


class ConvLayer(nn.Module):
    """
    Convolution operation on graphs
    """

    def __init__(self, atom_fea_len, nbr_fea_len):
        """
        Initialize ConvLayer. Message-level-embedder that sends message
        along the edges

        Parameters
        ----------

        atom_fea_len: int
          Number of atom hidden features.
        nbr_fea_len: int
          Number of bond features.
        """
        super(ConvLayer, self).__init__()
        self.atom_fea_len = atom_fea_len
        self.nbr_fea_len = nbr_fea_len
        # Create embedding for bond-atom-bond group
        self.fc_full = nn.Linear(
            2 * self.atom_fea_len + self.nbr_fea_len, 2 * self.atom_fea_len
        )
        # Create sigmoid for gate
        self.sigmoid = nn.Sigmoid()

        # Create softplus for core
        self.softplus1 = nn.Softplus()

        # Create normalization functions for the in-between steps
        self.bn1 = nn.BatchNorm1d(2 * self.atom_fea_len)
        self.bn2 = nn.BatchNorm1d(self.atom_fea_len)
        self.softplus2 = nn.Softplus()

    def forward(self, atom_in_fea, nbr_fea, nbr_fea_idx):
        """
        Forward pass

        N: Total number of atoms in the batch
        M: Max number of neighbors

        Parameters
        ----------

        atom_in_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom hidden features before convolution
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom

        Returns
        -------

        atom_out_fea: nn.Variable shape (N, atom_fea_len)
          Atom hidden features after convolution

        """

    def forward(self, atom_in_fea, nbr_fea, nbr_fea_idx):

        # atom_in_fea: (N, A)
        # nbr_fea_idx: (N, M)
        N, M = nbr_fea_idx.shape
        A = self.atom_fea_len
        B = self.nbr_fea_len

        # Gather neighbor atom features
        flat_idx = nbr_fea_idx.view(-1)  # (N*M,)

        assert (
            flat_idx.max().item() < N
        ), f"Found neighbor index {flat_idx.max().item()} ≥ number of atoms {N}"

        nbr_feats_flat = atom_in_fea[flat_idx]  # (N*M, A)
        atom_nbr_fea = nbr_feats_flat.view(N, M, A)  # (N, M, A)

        # Expand central atom feats
        atom_central = atom_in_fea[:N]
        atom_central = atom_central.unsqueeze(1)
        atom_central = atom_central.expand(N, M, A)

        # Concatenate central, neighbor, and bond features (atom-atom-edge)
        total_nbr_fea = torch.cat([atom_central, atom_nbr_fea, nbr_fea], dim=2)

        # Pass through Linear nn
        total_gated_fea = self.fc_full(total_nbr_fea)  # (N, M, 2A)

        # Normalize results. bn expects a 2D tensor. Flatten then reshape
        total_gated_fea = self.bn1(total_gated_fea.view(-1, A * 2)).view(N, M, A * 2)

        # Split into filter and core
        nbr_filter, nbr_core = total_gated_fea.chunk(2, dim=2)

        # Pass gate and core through sigmoid and softplus respectively
        nbr_filter = self.sigmoid(nbr_filter)
        nbr_core = self.softplus1(nbr_core)

        # Weight and sum over neighbors
        nbr_sumed = torch.sum(nbr_filter * nbr_core, dim=1)  # (N, A)

        # Apply normalization
        nbr_sumed = self.bn2(nbr_sumed)

        # Residual add + activation
        out = self.softplus2(atom_central[:, 0, :] + nbr_sumed)  # (N, A)

        return out


class CrystalGraphConvNet(nn.Module):
    """
    Create a crystal graph convolutional neural network for predicting
    predicting material properties (in this case, convergence)
    """

    def __init__(
        self,
        orig_atom_fea_len,
        nbr_fea_len,
        atom_fea_len=64,
        n_conv=3,
        h_fea_len=128,
        n_h=1,
    ):
        """
        Initialize CrystalGraphConvNet.

        Parameters
        ----------

        orig_atom_fea_len: int
          Number of atom features in the input.
        nbr_fea_len: int
          Number of bond features.
        atom_fea_len: int
          Number of hidden atom features in the convolutional layers
        n_conv: int
          Number of convolutional layers
        h_fea_len: int
          Number of hidden features after pooling
        n_h: int
          Number of hidden layers after pooling
        """
        super(CrystalGraphConvNet, self).__init__()

        # node-level embedder that brings raw inputs into same dimensionality for
        # convolution layer
        self.embedding = nn.Linear(orig_atom_fea_len, atom_fea_len)

        # Initialize convolutional stack. Each layer implements a round of message-passing
        self.convs = nn.ModuleList(
            [
                ConvLayer(atom_fea_len=atom_fea_len, nbr_fea_len=nbr_fea_len)
                for _ in range(n_conv)
            ]
        )

        # Map from conv_space to FC-space
        self.conv_to_fc = nn.Linear(atom_fea_len, h_fea_len)
        # add a smooth nonlinearity
        self.conv_to_fc_softplus = nn.Softplus()

        # Optional hidden layers. If you want more hidden layers after pooling,
        # created here

        if n_h > 1:
            self.fcs = nn.ModuleList(
                [nn.Linear(h_fea_len, h_fea_len) for _ in range(n_h - 1)]
            )

            self.softpluses = nn.ModuleList([nn.Softplus() for _ in range(n_h - 1)])

        # Finish processing the input with regression
        self.fc_out = nn.Linear(h_fea_len, 1)

    def forward(self, atom_fea, nbr_fea, nbr_fea_idx, crystal_atom_idx):
        """
        Forward pass

        N: Total number of atoms in the batch
        M: Max number of neighbors
        N0: Total number of crystals in the batch

        Parameters
        ----------

        atom_fea: Variable(torch.Tensor) shape (N, orig_atom_fea_len)
          Atom features from atom type
        nbr_fea: Variable(torch.Tensor) shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors
        nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx

        Returns
        -------

        prediction: nn.Variable shape (N, )
          Atom hidden features after convolution

        """

        # Embed atom feature vector
        atom_fea = self.embedding(atom_fea)

        # Pass atom embedding through convolution layer (message passing)
        for conv_func in self.convs:
            atom_fea = conv_func(atom_fea, nbr_fea, nbr_fea_idx)

        # Pool-to-crystal level features. For each crystal, gather its subset of N
        # atom features and sums them
        crys_fea = self.pooling(atom_fea, crystal_atom_idx)

        # Project conv-space to hidden MLP-space
        crys_fea = self.conv_to_fc(self.conv_to_fc_softplus(crys_fea))
        crys_fea = self.conv_to_fc_softplus(crys_fea)

        if hasattr(self, "fcs") and hasattr(self, "softpluses"):
            for fc, softplus in zip(self.fcs, self.softpluses):
                crys_fea = softplus(fc(crys_fea))

        out = self.fc_out(crys_fea)

        return out

    def pooling(self, atom_fea, crystal_atom_idx):
        """
        Pooling the atom features to crystal features

        N: Total number of atoms in the batch
        N0: Total number of crystals in the batch

        Parameters
        ----------

        atom_fea: Variable(torch.Tensor) shape (N, atom_fea_len)
          Atom feature vectors of the batch
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx
        """
        assert (
            sum([len(idx_map) for idx_map in crystal_atom_idx])
            == atom_fea.data.shape[0]
        )

        summed_fea = [
            torch.mean(atom_fea[idx_map], dim=0, keepdim=True)
            for idx_map in crystal_atom_idx
        ]

        return torch.cat(summed_fea, dim=0)


class ConvergenceRegressor(CrystalGraphConvNet):
    def __init__(
        self, orig_atom_fea_len, nbr_fea_len, atom_fea_len, n_conv, h_fea_len, n_h
    ):

        # Build as a regressor
        super().__init__(
            orig_atom_fea_len, nbr_fea_len, atom_fea_len, n_conv, h_fea_len, n_h
        )

        # Override the output to be 3 dim per atom. Pooling step unnecessary
        self.fc_out = nn.Linear(h_fea_len, 3)

    def forward(self, atom_fea, nbr_fea, nbr_fea_idx, crystal_atom_idx=None):

        # Embed atoms
        x = self.embedding(atom_fea)  # (N, F)

        # Message pass through convolution layers
        for conv in self.convs:
            x = conv(x, nbr_fea, nbr_fea_idx)  # still (N, F)

        # Project each atom into the 3-vector space
        x = self.conv_to_fc_softplus(self.conv_to_fc(x))  # (N, H)

        # If extra layers are created, run through
        if hasattr(self, "fcs"):
            for fc, act in zip(self.fcs, self.softpluses):
                # applies linear layer to x and applies soft max. Repeats
                x = act(fc(x))

        # Serves as our per-atom displacement vectors
        out = self.fc_out(x)

        return out



import csv
import json
import os
import random
import warnings

import numpy as np
import torch
from pymatgen.core.structure import Structure
from torch.utils.data import Dataset
from torch.utils.data.sampler import SubsetRandomSampler
from torch.utils.data import DataLoader


def get_train_val_test_loader(
    dataset,
    batch_size=64,
    train_ratio=None,
    val_ratio=0.1,
    test_ratio=0.1,
    return_test=False,
    num_workers=1,
    pin_memory=False,
):
    """
    Utility function for dividing a dataset to train, val, test datasets.
    """
    total_size = len(dataset)
    if train_ratio is None:
        assert val_ratio + test_ratio < 1
        train_ratio = 1 - val_ratio - test_ratio
        warnings.warn(
            f"train_ratio is None, using 1 - val_ratio - test_ratio = {train_ratio} as training data."
        )
    else:
        assert train_ratio + val_ratio + test_ratio <= 1

    indices = list(range(total_size))
    train_size = int(train_ratio * total_size)
    test_size = int(test_ratio * total_size)
    valid_size = int(val_ratio * total_size)

    train_sampler = SubsetRandomSampler(indices[:train_size])
    val_sampler = SubsetRandomSampler(indices[train_size : train_size + valid_size])
    if return_test:
        test_sampler = SubsetRandomSampler(
            indices[train_size + valid_size : train_size + valid_size + test_size]
        )

    train_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        num_workers=num_workers,
        collate_fn=collate_pool,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=val_sampler,
        num_workers=num_workers,
        collate_fn=collate_pool,
        pin_memory=pin_memory,
    )
    if return_test:
        test_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            sampler=test_sampler,
            num_workers=num_workers,
            collate_fn=collate_pool,
            pin_memory=pin_memory,
        )
        return train_loader, val_loader, test_loader
    return train_loader, val_loader, None


def collate_pool(dataset_list):
    """
    Collate a list of data points for per-atom displacement training:
      ((atom_fea, nbr_fea, nbr_idx), dr_true, struct_relaxed)
    Returns:
      (atom_batch, nbr_batch, idx_batch, crystal_atom_idx), dr_batch, batch_struct
    """
    batch_atom_fea, batch_nbr_fea, batch_nbr_idx = [], [], []
    crystal_atom_idx, batch_dr, batch_struct = [], [], []
    base_idx = 0
    for (atom_fea, nbr_fea, nbr_idx), dr_true, struct_relaxed in dataset_list:
        n_i = atom_fea.shape[0]
        batch_atom_fea.append(atom_fea)
        batch_nbr_fea.append(nbr_fea)
        # shift neighbor indices by base_idx for batching
        batch_nbr_idx.append(nbr_idx + base_idx)
        # record this crystal's atom indices
        crystal_atom_idx.append(
            torch.arange(base_idx, base_idx + n_i, dtype=torch.long)
        )
        batch_dr.append(dr_true)
        batch_struct.append(struct_relaxed)
        base_idx += n_i

    atom_batch = torch.cat(batch_atom_fea, dim=0)
    nbr_batch = torch.cat(batch_nbr_fea, dim=0)
    idx_batch = torch.cat(batch_nbr_idx, dim=0)
    dr_batch = torch.cat(batch_dr, dim=0)

    return (atom_batch, nbr_batch, idx_batch, crystal_atom_idx), dr_batch, batch_struct


class GaussianDistance(object):
    """Gaussian smearing for bond distances"""

    def __init__(self, dmin: float, dmax: float, step: float, var=None):
        assert dmin < dmax
        assert dmax - dmin > step
        self.filter = np.arange(dmin, dmax + step, step)
        self.var = var if var is not None else step

    def expand(self, distances: np.ndarray):
        return np.exp(-((distances[..., np.newaxis] - self.filter) ** 2) / self.var**2)


class AtomInitializer(object):
    """Base class for atom type embeddings"""

    def __init__(self, atom_types):
        self.atom_types: set = set(atom_types)
        self._embedding: dict = {}

    def load_state_dict(self, state_dict):
        self._embedding = state_dict
        self.atom_types = set(self._embedding.keys())
        self._decodedict = {v: k for k, v in self._embedding.items()}

    def get_atom_fea(self, atom_type):
        assert atom_type in self.atom_types
        return self._embedding[atom_type]

    def state_dict(self):
        return self._embedding

    def decode(self, idx):
        if not hasattr(self, "_decodedict"):
            self._decodedict = {v: k for k, v in self._embedding.items()}
        return self._decodedict[idx]


class AtomCustomJSONInitializer(AtomInitializer):
    """Load element embeddings from JSON"""

    def __init__(self, elem_embedding_file):
        with open(elem_embedding_file) as f:
            elem_embedding: dict = json.load(f)
        elem_embedding = {int(k): v for k, v in elem_embedding.items()}
        super().__init__(set(elem_embedding.keys()))
        for key, value in elem_embedding.items():
            self._embedding[key] = np.array(value, dtype=float)


class CIFData(Dataset):
    """
    CIFData with on-the-fly "relaxed" structures via random perturbation.
    Returns original features, per-atom displacement targets, and perturbed Structures.
    """

    def __init__(
        self,
        root_dir,
        max_num_nbr=12,
        radius=8.0,
        dmin=0.0,
        step=0.2,
        random_seed=123,
        perturb_std=0.01,
    ):
        self.root_dir = root_dir
        self.max_num_nbr = max_num_nbr
        self.radius = radius
        self.dmin = dmin
        self.step = step
        self.perturb_std = perturb_std
        random.seed(random_seed)

        # read id_prop.csv
        id_prop_file = os.path.join(root_dir, "id_prop.csv")
        assert os.path.exists(id_prop_file), "id_prop.csv missing"
        with open(id_prop_file) as f:
            reader = csv.reader(f)
            self.id_prop_data = list(reader)
        random.shuffle(self.id_prop_data)

        # load embeddings and distance filter
        atom_init_file = os.path.join(root_dir, "atom_init.json")
        assert os.path.exists(atom_init_file), "atom_init.json missing"
        self.ari = AtomCustomJSONInitializer(atom_init_file)
        self.gdf = GaussianDistance(dmin=dmin, dmax=radius, step=step)

    def __len__(self):
        return len(self.id_prop_data)

    def __getitem__(self, idx):
        cif_id, _, _ = self.id_prop_data[idx]
        # load original crystal
        crystal = Structure.from_file(os.path.join(self.root_dir, f"{cif_id}.cif"))

        # original cartesian positions
        orig_pos = np.array(crystal.cart_coords)
        # generate small Gaussian perturbation
        delta = np.random.normal(scale=self.perturb_std, size=orig_pos.shape)
        new_coords = orig_pos + delta

        # build perturbed structure
        struct_relaxed = Structure(
            crystal.lattice, crystal.species, new_coords, coords_are_cartesian=True
        )

        # per-atom displacement target
        dr_true = torch.tensor(delta, dtype=torch.float32)

        # featurize original crystal
        atom_fea, nbr_fea, nbr_idx = self._featurize(crystal)

        return (atom_fea, nbr_fea, nbr_idx), dr_true, struct_relaxed

    def _featurize(self, crystal):
        # atom features
        atom_feats = [self.ari.get_atom_fea(site.specie.number) for site in crystal]
        atom_fea = torch.tensor(np.vstack(atom_feats), dtype=torch.float32)

        # build neighbor list
        all_nbrs = crystal.get_all_neighbors(self.radius, include_index=True)
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]

        nbr_idx_list, nbr_dist_list = [], []
        for nbr in all_nbrs:
            if len(nbr) < self.max_num_nbr:
                warnings.warn("Neighbors < max_num_nbr, padding with zeros.")
                idxs = [x[2] for x in nbr] + [0] * (self.max_num_nbr - len(nbr))
                dists = [x[1] for x in nbr] + [self.radius + 1] * (
                    self.max_num_nbr - len(nbr)
                )
            else:
                idxs = [x[2] for x in nbr[: self.max_num_nbr]]
                dists = [x[1] for x in nbr[: self.max_num_nbr]]
            nbr_idx_list.append(idxs)
            nbr_dist_list.append(dists)

        nbr_idx_arr = np.array(nbr_idx_list)
        nbr_dist_arr = np.array(nbr_dist_list)
        nbr_fea = self.gdf.expand(nbr_dist_arr)

        nbr_idx = torch.tensor(nbr_idx_arr, dtype=torch.long)
        nbr_fea = torch.tensor(nbr_fea, dtype=torch.float32)

        return atom_fea, nbr_fea, nbr_idx



from groovemat.cli import cli, train, predict
from groovemat.data import (
    AtomCustomJSONInitializer,
    AtomInitializer,
    GaussianDistance,
    CIFData,
)
from groovemat.model import CrystalGraphConvNet, ConvLayer
from groovemat.matgl_loss import MatGLLoss

__all__ = [
    "cli",
    "train",
    "predict",
    "AtomCustomJSONInitializer",
    "AtomInitializer",
    "GaussianDistance",
    "CIFData",
    "CrystalGraphConvNet",
    "ConvLayer",
    "MatGLLoss",
]



import torch


class Normalizer(object):
    """Normalize a Tensor and restore it later."""

    def __init__(self, tensor):
        """tensor is taken as a sample to calculate the mean and std"""
        self.mean = torch.mean(tensor)
        self.std = torch.std(tensor)

    def norm(self, tensor):
        return (tensor - self.mean) / self.std

    def denorm(self, normed_tensor):
        return normed_tensor * self.std + self.mean

    def state_dict(self):
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict):
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]



import os
import shutil
import sys
import time
import warnings
from random import sample

import numpy as np
import torch
import torch.optim as optim
from sklearn import metrics
from torch.optim.lr_scheduler import MultiStepLR
import torch.nn as nn
import matplotlib.pyplot as plt
from pymatgen.core import Structure
import click

from groovemat.data import CIFData
from groovemat.data import collate_pool, get_train_val_test_loader
from groovemat.model import ConvergenceRegressor
from groovemat.matgl_loss import MatGLLoss
from groovemat.utils.normalizer import Normalizer

BANNER = r"""
+----------------------------------------------------------+
|      ____                          __  __       _        |
|     / ___|_ __ ___   _____   _____|  \/  | __ _| |_      |
|    | |  _| '__/ _ \ / _ \ \ / / _ \ |\/| |/ _` | __|     |
|    | |_| | | | (_) | (_) \ V /  __/ |  | | (_| | |_      |
|     \____|_|  \___/ \___/ \_/ \___|_|  |_|\__,_|\__|     |
|                                                          |
|                Get those atoms moving!                   |
+----------------------------------------------------------+
"""
args = None


class CustomGroup(click.Group):
    def get_help(self, ctx):
        base_help = super().get_help(ctx)
        return BANNER + "\n\n" + base_help


@click.group(cls=CustomGroup, invoke_without_command=True)
@click.pass_context
def cli(ctx):
    """
    GrooveMat command‑line interface.
    """
    if ctx.invoked_subcommand is None:
        click.echo(cli.get_help(ctx))


@cli.command()
@click.argument("checkpoint", type=click.Path(exists=True))
@click.argument("input_cif", type=click.Path(exists=True))
@click.argument("output_cif", type=click.Path())
@click.option("--cuda/--no-cuda", default=False, help="Enable CUDA if available")
def predict(checkpoint, input_cif, output_cif, cuda):
    """
    Take a trained model at CHECKPOINT, relax INPUT_CIF,
    and write the relaxed structure to OUTPUT_CIF.
    """
    ckpt = torch.load(checkpoint, map_location="cpu")
    saved_args = ckpt["args"]
    state_dict = ckpt["state_dict"]

    device = torch.device("cuda" if cuda and torch.cuda.is_available() else "cpu")
    model = ConvergenceRegressor(
        atom_fea_len=saved_args.atom_fea_len,
        n_conv=saved_args.n_conv,
        h_fea_len=saved_args.h_fea_len,
        n_h=saved_args.n_h,
    )
    model.load_state_dict(state_dict)
    model.to(device).eval()

    normalizer = Normalizer()

    struct = Structure.from_file(input_cif)

    ds = CIFData(root_dir=None, file_list=[input_cif], perturb_std=0.0)
    loader = get_train_val_test_loader(
        dataset=ds,
        batch_size=1,
        train_ratio=None,
        val_ratio=None,
        test_ratio=None,
        return_test=True,
        num_workers=0,
    )[
        -1
    ]  # test_loader

    with torch.no_grad():
        for inputs, _dr_true, batch_struct in loader:
            atom_fea, nbr_fea, nbr_idx, idx_map = inputs
            atom_fea = atom_fea.to(device)
            nbr_fea = nbr_fea.to(device)
            nbr_idx = nbr_idx.to(device)

            pred_dr_n = model(atom_fea, nbr_fea, nbr_idx)
            pred_dr_n = normalizer.denorm(pred_dr_n)
            pred_dr_n = pred_dr_n.cpu().numpy().reshape(-1, 3)

    orig_coords = np.array(struct.cart_coords)
    new_coords = orig_coords + pred_dr_n

    relaxed = Structure(
        lattice=struct.lattice,
        species=struct.species,
        coords=new_coords,
        coords_are_cartesian=True,
    )

    relaxed.to(filename=output_cif, fmt="cif")
    click.echo(f"Wrote relaxed structure ➔ {output_cif}")


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("data_options", nargs=-1, type=click.STRING)
@click.option("--disable-cuda", is_flag=True, help="Disable CUDA")
@click.option(
    "-j",
    "--workers",
    default=0,
    type=int,
    metavar="N",
    help="number of data loading workers (default: 0)",
)
@click.option(
    "--epochs",
    default=30,
    type=int,
    metavar="N",
    help="number of total epochs to run (default: 30)",
)
@click.option(
    "--start-epoch",
    default=0,
    type=int,
    metavar="N",
    help="manual epoch number (useful on restarts)",
)
@click.option(
    "-b",
    "--batch-size",
    default=256,
    type=int,
    metavar="N",
    help="mini‑batch size (default: 256)",
)
@click.option(
    "--lr",
    "--learning-rate",
    "lr",
    default=0.01,
    type=float,
    metavar="LR",
    help="initial learning rate (default: 0.01)",
)
@click.option(
    "--lr-milestones",
    default=[100],
    multiple=True,
    type=int,
    metavar="N",
    help="milestones for scheduler (default: [100])",
)
@click.option("--momentum", default=0.9, type=float, metavar="M", help="momentum")
@click.option(
    "--weight-decay",
    "--wd",
    "weight_decay",
    default=0.0,
    type=float,
    metavar="W",
    help="weight decay (default: 0)",
)
@click.option(
    "-p",
    "--print-freq",
    "print_freq",
    default=10,
    type=int,
    metavar="N",
    help="print frequency (default: 10)",
)
@click.option(
    "--resume",
    default="",
    type=click.Path(),
    metavar="PATH",
    help="path to latest checkpoint (default: none)",
)
@click.option(
    "--train-ratio",
    type=float,
    default=None,
    help="number of training data to be loaded (mutually exclusive with --train-size)",
)
@click.option(
    "--train-size",
    type=int,
    default=None,
    help="number of training data to be loaded (mutually exclusive with --train-ratio)",
)
@click.option(
    "--val-ratio",
    type=float,
    default=0.1,
    help="percentage of validation data to be loaded (default: 0.1)",
)
@click.option(
    "--val-size",
    type=int,
    default=None,
    help="number of validation data to be loaded (mutually exclusive with --val-ratio)",
)
@click.option(
    "--test-ratio",
    type=float,
    default=0.1,
    help="percentage of test data to be loaded (default: 0.1)",
)
@click.option(
    "--test-size",
    type=int,
    default=None,
    help="number of test data to be loaded (mutually exclusive with --test-ratio)",
)
@click.option(
    "--optim",
    default="SGD",
    type=click.Choice(["SGD", "Adam"]),
    metavar="SGD",
    help="choose an optimizer, SGD or Adam (default: SGD)",
)
@click.option(
    "--atom-fea-len",
    default=64,
    type=int,
    metavar="N",
    help="number of hidden atom features in conv layers",
)
@click.option(
    "--h-fea-len",
    default=128,
    type=int,
    metavar="N",
    help="number of hidden features after pooling",
)
@click.option(
    "--n-conv", default=3, type=int, metavar="N", help="number of conv layers"
)
@click.option(
    "--n-h",
    default=1,
    type=int,
    metavar="N",
    help="number of hidden layers after pooling",
)
@click.option(
    "--perturb-std",
    default=0.01,
    type=float,
    help="standard deviation for random displacement (Å)",
)
def train(
    data_options,
    disable_cuda,
    workers,
    epochs,
    start_epoch,
    batch_size,
    lr,
    lr_milestones,
    momentum,
    weight_decay,
    print_freq,
    resume,
    train_ratio,
    train_size,
    val_ratio,
    val_size,
    test_ratio,
    test_size,
    optim,
    atom_fea_len,
    h_fea_len,
    n_conv,
    n_h,
    perturb_std,
):
    """
    Train the GrooveMat model
    """

    class Args:
        pass

    global args
    args = Args()
    args.data_options = data_options
    args.disable_cuda = disable_cuda
    args.workers = workers
    args.epochs = epochs
    args.start_epoch = start_epoch
    args.batch_size = batch_size
    args.lr = lr
    args.lr_milestones = list(lr_milestones)
    args.momentum = momentum
    args.weight_decay = weight_decay
    args.print_freq = print_freq
    args.resume = resume
    args.train_ratio = train_ratio
    args.train_size = train_size
    args.val_ratio = val_ratio
    args.val_size = val_size
    args.test_ratio = test_ratio
    args.test_size = test_size
    args.optim = optim
    args.atom_fea_len = atom_fea_len
    args.h_fea_len = h_fea_len
    args.n_conv = n_conv
    args.n_h = n_h
    args.perturb_std = perturb_std

    args.cuda = not args.disable_cuda and torch.cuda.is_available()

    # Load data
    dataset = CIFData(*args.data_options)
    train_loader, val_loader, test_loader = get_train_val_test_loader(
        dataset=dataset,
        batch_size=args.batch_size,
        train_ratio=args.train_ratio,
        num_workers=args.workers,
        val_ratio=args.val_ratio,
        test_ratio=args.test_ratio,
        pin_memory=args.cuda,
        return_test=True
    )

    if len(dataset) < 500:
        warnings.warn(
            "Dataset has less than 500 data points. " "Lower accuracy is expected. "
        )
        sample_data_list = [dataset[i] for i in range(len(dataset))]

    else:
        sample_data_list = [dataset[i] for i in sample(range(len(dataset)), 500)]

    _, sample_target, _ = collate_pool(sample_data_list)
    normalizer = Normalizer(sample_target)

    # Build model
    structures, _, _ = dataset[0]
    orig_atom_fea_len = structures[0].shape[-1]
    nbr_fea_len = structures[1].shape[-1]

    # after you compute orig_atom_fea_len, nbr_fea_len:
    device = torch.device("cuda" if args.cuda else "cpu")

    model = ConvergenceRegressor(
        orig_atom_fea_len,
        nbr_fea_len,
        atom_fea_len=args.atom_fea_len,
        n_conv=args.n_conv,
        h_fea_len=args.h_fea_len,
        n_h=args.n_h,
    ).to(device)

    if args.cuda:
        model.cuda()

    # Grab the loss function
    loss_fn = MatGLLoss(model_name="M3GNet-MP-2021.2.8-PES")
    criterion = nn.MSELoss()

    # Choose optimizer
    if args.optim == "SGD":
        optimizer = torch.optim.SGD(
            model.parameters(),
            args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay,
        )

    elif args.optim == "Adam":
        optimizer = torch.optim.Adam(
            model.parameters(), args.lr, weight_decay=args.weight_decay
        )

    else:
        raise NameError("Only SGD or Adam is allowed as --optim")

    # Optionally resume from checkpoint
    best_m3g_error = 1000
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            checkpoint = torch.load(args.resume)
            args.start_epoch = checkpoint["epoch"]
            best_m3g_error = checkpoint["best_m3g_error"]
            model.load_state_dict(checkpoint["state_dict"])
            optimizer.load_state_dict(checkpoint["optimizer"])
            normalizer.load_state_dict(checkpoint["normalizer"])
            print(
                "=> loaded checkpoint '{}' (epoch {})".format(
                    args.resume, checkpoint["epoch"]
                )
            )
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    # Initialize scheduler, in charge of updating the learning rate
    scheduler = MultiStepLR(optimizer, milestones=args.lr_milestones, gamma=0.1)

    train_losses = []
    val_losses = []
    train_maes = []
    val_mae_losses = []

    for epoch in range(args.start_epoch, args.epochs):

        # Train for one epoch
        train_loss, train_mae = train_helper(
            normalizer,  # your Normalizer
            train_loader,  # the DataLoader
            model,  # your model (already on device)
            loss_fn,  # the MatGLLoss instance
            criterion,  # the pointwise loss (MSELoss)
            optimizer,  # the optimizer
            device,  # the torch.device
            epoch,  # current epoch index
        )

        # evaluate on validation set
        m3g_error, mae_val_error = validate(
            val_loader, model, criterion, normalizer, device
        )

        train_losses.append(train_loss)
        train_maes.append(train_mae)
        val_losses.append(m3g_error)
        val_mae_losses.append(mae_val_error)

        if m3g_error != m3g_error:
            print("Exit due to NaN")
            sys.exit(1)

        scheduler.step()

        # Remember the best m3g_eror and save checkpoint
        best_m3g_error = min(m3g_error, best_m3g_error)
        is_best = m3g_error < best_m3g_error
        save_checkpoint(
            {
                "epoch": epoch + 1,
                "state_dict": model.state_dict(),
                "best_m3g_error": best_m3g_error,
                "optimizer": optimizer.state_dict(),
                "normalizer": normalizer.state_dict(),
                "args": vars(args),
            },
            is_best,
        )

    print("---------Evaluate Model on Test Set---------------")
    best_checkpoint = torch.load("model_best.pth.tar")
    model.load_state_dict(best_checkpoint["state_dict"])
    validate(test_loader, model, criterion, normalizer, device, test=True)
    # Test best model
    plt.figure()
    plt.plot(range(1, args.epochs + 1), train_maes, label="Train MAE (Å)")
    plt.plot(range(1, args.epochs + 1), val_mae_losses, label="Val MAE   (Å)")
    plt.xlabel("Epoch")
    plt.ylabel("MAE (Å)")
    plt.title("Training & Validation MAE")
    plt.legend()
    plt.grid(True)
    plt.savefig("mae_vs_epoch.png", dpi=300)


def train_helper(
    normalizer: Normalizer,
    train_loader,
    model: nn.Module,
    loss_fn: MatGLLoss,
    criterion: nn.Module,
    optimizer: torch.optim.Optimizer,
    device: torch.device,
    epoch: int,
):
    model.train()

    train_meter = AverageMeter()
    mae_meter = AverageMeter()

    for i, (inputs, dr_true, batch_struct) in enumerate(train_loader):
        atom_fea, nbr_fea, nbr_idx, crystal_atom_idx = inputs
        atom_fea = atom_fea.to(device)
        nbr_fea = nbr_fea.to(device)
        nbr_idx = nbr_idx.to(device)

        # forward over the full batch → (N_total, 3)
        pred_dr_n = model(atom_fea, nbr_fea, nbr_idx)

        # split the true Δr into per‑crystal chunks
        sizes = [len(idx_map) for idx_map in crystal_atom_idx]
        dr_list = dr_true.split(sizes, dim=0)  # list of (n_i,3)

        # accumulate MatGL loss per crystal
        total_loss = 0.0
        for idx_map, struct_relaxed, dr_i in zip(
            crystal_atom_idx, batch_struct, dr_list
        ):
            # extract this crystal’s atom predictions
            pred_i = pred_dr_n[idx_map]  # (n_i, 3)
            x_flat = pred_i.view(-1)  # (3*n_i,)

            # reconstruct the initial (perturbed) coords:
            #   r_initial = r_relaxed - Δr_true
            coords_relaxed = np.array(struct_relaxed.cart_coords)
            delta_np = dr_i.cpu().numpy()  # (n_i,3)
            coords_initial = coords_relaxed - delta_np

            # build a new Structure at the perturbed geometry
            struct_initial = Structure(
                struct_relaxed.lattice,
                struct_relaxed.species,
                coords_initial,
                coords_are_cartesian=True,
            )

            # use the perturbed Structure for the loss
            total_loss += loss_fn(x_flat, struct_initial, classifier=criterion)

        # average over the batch
        loss = total_loss / len(batch_struct)

        train_meter.update(loss.item(), len(batch_struct))

        optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
        optimizer.step()

        with torch.no_grad():
            # denormalize the whole‑batch predictions
            pred_all = normalizer.denorm(pred_dr_n)  # (N_total, 3)
            dr_all = dr_true  # (N_total, 3)

            # split into per‑crystal chunks
            sizes = [len(idx_map) for idx_map in crystal_atom_idx]
            pred_list = pred_all.split(sizes, dim=0)  # list of (n_i,3)
            dr_list = dr_all.split(sizes, dim=0)  # list of (n_i,3)

            # compute each crystal’s ⟨|change in r_pred – change in r_true|⟩
            errs = [(p - d).abs().mean() for p, d in zip(pred_list, dr_list)]
            mean_err = torch.stack(errs).mean().item()
            batch_mae = (pred_all - dr_true).abs().mean().item()

        mae_meter.update(batch_mae, dr_true.size(0))

        if i % args.print_freq == 0:
            print(
                f"Epoch {epoch} | Iter {i}: "
                f"train loss={loss.item():.4f}, "
                f"⟨|Δr|⟩={mean_err:.3f} Å"
            )

    return train_meter.avg, mae_meter.avg


def validate(val_loader, model, criterion, normalizer: Normalizer, device, test=False):
    batch_time = AverageMeter()
    m3g_errors = AverageMeter()
    mae_errors = AverageMeter()

    # If we're in a “test” run, collect predictions/ids for csv
    if test:
        test_preds = []
        test_targets = []
        test_cif_ids = []

    model.eval()
    start = time.time()

    with torch.no_grad():
        for i, (inputs, dr_true, batch_cif_ids) in enumerate(val_loader):
            # unpack your graph inputs
            atom_fea, nbr_fea, nbr_idx, _ = inputs

            # move everything onto device
            atom_fea = atom_fea.to(device)
            nbr_fea = nbr_fea.to(device)
            nbr_idx = nbr_idx.to(device)
            dr_true = dr_true.to(device)

            # normalize target displacement
            dr_true_n = normalizer.norm(dr_true)

            # forward + MSE loss
            pred_n = model(atom_fea, nbr_fea, nbr_idx)
            loss = criterion(pred_n, dr_true_n)

            # record MSE loss
            m3g_errors.update(loss.item(), dr_true.size(0))

            # denormalize for MAE
            pred = normalizer.denorm(pred_n)
            m3g = (pred - dr_true).abs().mean().item()
            # record MAE per atom, not MatGLLoss instance
            mae_errors.update(m3g, dr_true.size(0))

            # if in test mode, stash for CSV
            if test:
                test_preds += pred.view(-1).cpu().tolist()
                test_targets += dr_true.view(-1).cpu().tolist()
                test_cif_ids += batch_cif_ids

            # timing
            batch_time.update(time.time() - start)
            start = time.time()

            # progress print
            if i % args.print_freq == 0:
                print(
                    f"Val: [{i}/{len(val_loader)}]  "
                    f"Time {batch_time.val:.3f} ({batch_time.avg:.3f})  "
                    f"Loss {m3g_errors.val:.4f} ({m3g_errors.avg:.4f})  "
                    f"M3g Error {m3g_errors.val:.3f} ({m3g_errors.avg:.3f})"
                )

    # if test, dump CSV
    if test:
        import csv

        with open("test_results.csv", "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["cif_id", "true_Δr", "pred_Δr"])
            for cid, t, p in zip(test_cif_ids, test_targets, test_preds):
                writer.writerow((cid, t, p))

    # Final Summary
    star = "**" if test else "*"
    print(f" {star} M3g Error Avg {m3g_errors.avg:.3f}")
    print(f"{star}  Val MAE Avg = {mae_errors.avg:.4f} Å")
    return m3g_errors.avg, mae_errors.avg


def class_eval(prediction, target: torch.Tensor):
    prediction = np.exp(prediction.numpy())
    target = target.numpy()
    pred_label = np.argmax(prediction, axis=1)
    target_label = np.squeeze(target)
    if not target_label.shape:
        target_label = np.asarray([target_label])
    if prediction.shape[1] == 2:
        precision, recall, fscore, _ = metrics.precision_recall_fscore_support(
            target_label, pred_label, average="binary"
        )
        auc_score = metrics.roc_auc_score(target_label, prediction[:, 1])
        accuracy = metrics.accuracy_score(target_label, pred_label)
    else:
        raise NotImplementedError
    return accuracy, precision, recall, fscore, auc_score


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def save_checkpoint(state, is_best, filename="checkpoint.pth.tar"):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, "model_best.pth.tar")


def adjust_learning_rate(optimizer, epoch, k):
    """Sets the learning rate to the initial LR decayed by 10 every k epochs"""
    assert type(k) is int
    lr = args.lr * (0.1 ** (epoch // k))
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


if __name__ == "__main__":
    main()
